import pandas as pd
import matplotlib.pyplot as plt
import os
import statsmodels.stats.proportion as sp


def category_churn_cohorts(data_set_path, cat_col):
    churn_data = prepare_category_data(data_set_path,cat_col)
    summary = category_churn_summary(churn_data,cat_col,data_set_path)
    category_churn_plot(cat_col, summary, data_set_path)

def prepare_category_data(data_set_path,cat_col):
    assert os.path.isfile(data_set_path),'"{}" is not a valid dataset path'.format(data_set_path)
    churn_data = pd.read_csv(data_set_path,index_col=[0,1])
    churn_data[cat_col].fillna('-na-',inplace=True)
    return churn_data

def category_churn_summary(churn_data, cat_col,data_set_path):
    summary = churn_data.groupby(cat_col).agg({cat_col:'count','is_churn': ['sum','mean']})
    intervals = sp.proportion_confint(summary['is_churn','sum'],summary[ (cat_col,'count')],method='wilson')
    summary[cat_col + '_percent'] = (1.0/churn_data.shape[0]) * summary[(cat_col,'count')]
    summary['lo_conf'] = intervals[0]
    summary['hi_conf'] = intervals[1]
    summary['lo_int'] = summary[('is_churn','mean')]-summary['lo_conf']
    summary['hi_int'] = summary['hi_conf'] - summary[('is_churn','mean')]
    save_path = data_set_path.replace('.csv', '_' + cat_col + '_churn_category.csv')
    summary.to_csv(save_path)
    return summary


def category_churn_plot(cat_col, summary,data_set_path):
    n_category = summary.shape[0]
    plt.figure(figsize=(max(4,.5*n_category), 4))
    plt.bar(x=summary.index,height=summary[('is_churn','mean')],
            yerr=summary[['lo_int','hi_int']].transpose().values,capsize=80/n_category)
    plt.xlabel('Category Average of Churn for  "%s"' % cat_col)
    plt.ylabel('Category Churn Rate')
    plt.grid()
    save_path = data_set_path.replace('.csv', '_' + cat_col + '_churn_category.png')
    plt.savefig(save_path)
    print('Saving plot to %s' % save_path)




from listing_5_2_dataset_stats import dataset_stats
from listing_7_5_fat_tail_scores import fat_tail_scores
from listing_6_4_find_metric_groups import find_metric_groups
from listing_6_3_apply_metric_groups import apply_metric_groups
from listing_6_5_ordered_correlation_matrix import ordered_correlation_matrix

def prepare_data(data_set_path,group_corr_thresh=0.55):
    dataset_stats(data_set_path)
    fat_tail_scores(data_set_path)
    find_metric_groups(data_set_path,group_corr_thresh)
    apply_metric_groups(data_set_path)
    ordered_correlation_matrix(data_set_path)



from postgres import Postgres
import json
import pandas as pd
from importlib import import_module
from copy import copy
import re
import os
import sys
import argparse

"""
####################################################################################################
Arguments
"""
parser = argparse.ArgumentParser()
# Run control arguments
parser.add_argument("--schema", type=str, help="The name of the schema", default='socialnet7')
parser.add_argument("--chapter", type=int, help="The chapter of the listing", default=2)
parser.add_argument("--listing", nargs='*', type=int, help="The number of the listing", default=[1])
parser.add_argument("--insert", action="store_true", default=False,help="Use the insert version of a metric SQL, if available")
parser.add_argument("--version", nargs='*', help="Alternative listing _parameter_ verions (optional)",default=[])


"""
####################################################################################################
Constants
"""

pd.set_option('display.max_columns', 500)
pd.set_option('display.width', 1000)
print_num_rows=10

reserved_param_keywords = ('listing', 'mode','type','schema','name','chapter','full_name')

"""
####################################################################################################
Functions
"""

def _full_listing_name(chapter, listing, name, insert=False):
    """
    Creates the name of a listing file from the components.
    The names are like "listing_C_N_<name>" where C is the chapter, N is the number and
    <name> is the listing name.  For example, "listing_2_1_net_retention".
    "insert" listings are a special SQL listing type used in chapter 7.
    :param chapter:
    :param listing:
    :param name:
    :param insert:
    :return:
    """
    prefix = 'listing' if not insert else 'insert'
    full_name = f"{prefix}_{chapter}_{listing}_{name}"
    return full_name


def _sql_listing_from_params(param_dict,insert):
    """
    Utility to take parameters in a dictionary loaded from configuration, and call the sql_listing function
    Pops those dictionary parameters that are parameters of the sql_listing parameters, and passes what is
    left as the bind variables
    :param param_dict:
    :param insert:
    :return:
    """
    chapter = param_dict.pop('chapter')
    listing = param_dict.pop('listing')
    name = param_dict.pop('name')
    schema = param_dict.pop('schema')
    mode = param_dict.pop('mode')
    save_ext = param_dict.pop('save_ext') if 'save_ext' in param_dict else None
    sql_listing(chapter, listing, name, schema, mode, param_dict, insert=insert, save_ext=save_ext)


def sql_listing(chapter, listing, name, schema, mode, param_dict, insert=False, save_ext=None):
    """

    Run a SQL listing.  The sql file is loaded, and then any non-reserved keyword in the parameters is treated as a
    string to be replaced in the sql string. The SQL is then printed out.

    :param chapter:
    :param listing:
    :param name:
    :param schema:
    :param mode:The allowed modes are:
        run : The SQL returns no result
        one : The SQL should return one row result to be printed
        top : The SQL returns many results, print the top N (given by global print_num_rows)
    :param param_dict: the bind variables as a dictionary
    :param insert: flag to use the insert form of a query; see chapter 7
    :param save_ext:
    :return:
    """

    with open('../../listings/chap%d/%s.sql' % (chapter, _full_listing_name(chapter, listing, name, insert)), 'r') as myfile:

        db = Postgres("postgres://%s:%s@localhost/%s" % (os.environ['CHURN_DB_USER'],os.environ['CHURN_DB_PASS'],os.environ['CHURN_DB']))

        # prefix the search path onto the listing, which does not specify the schema
        sql = "set search_path = '%s'; " % schema

        # load the sql file
        sql = sql + myfile.read()

        # bind any parameters that match strings in the sql
        param_keys = [p for p in param_dict.keys() if p not in reserved_param_keywords]
        for p in param_keys:
            sql = sql.replace(p, str(param_dict[p]))

        # Print the sql (then remove the newlines)
        print('SQL:\n----------\n'+sql+'\n----------\nRESULT:')
        sql = sql.replace('\n', ' ')

        # Run in the manner indicated by the mode
        if  mode  == 'run':
            db.run(sql)
        elif  mode  == 'one':
            res = db.one(sql)
            print(res)
        elif  mode  == 'top' or mode == 'save':
            res = db.all(sql)
            df = pd.DataFrame(res)
            if  mode  == 'save':
                save_path = '../../../fight-churn-output/' + schema+ '/'
                os.makedirs(save_path,exist_ok=True)
                csv_path= save_path + schema+ '_' + name.replace(
                    'listing_{}_{}_'.format(chapter,listing),'')
                if save_ext:
                    csv_path = csv_path + '_' + save_ext
                csv_path = csv_path + '.csv'
                print('Saving: %s' % csv_path)
                df.to_csv(csv_path, index=False)
            else:
                print(df.head(print_num_rows))
        else:
            print('Unknown run mode for SQL example')
            exit(-4)

def _python_listing_from_params(param_dict):
    """
    Utility to call the python_listing function from a parameter dictionary, loaded from the configuration
    Pops the parameters that are part of the function signature, and passes the remaining parameters as dictionary
    of function parameters
    :param param_dict:
    :return:
    """
    chapter = param_dict.pop('chapter')
    listing = param_dict.pop('listing')
    name = param_dict.pop('name')

    python_listing(chapter, listing, name, param_dict)


def python_listing(chapter, listing, name, param_dict):
    """
    Runs one python code listing.  Each listing is defined as a model, and should have one function in it, which is the name of
    the listing file, without the part saying listing_X_Y.  If the function is found in the module it is run; all of the
    non-reserved keywords in the parameter dictionary are treated as the parameters
    :param chapter:
    :param listing:
    :param name:
    :param param_dict: dictionary produced by load_and_check_listing_params
    :return:
    """

    # make a dictionary of just the non-reserved keywords, which should be the parameters
    example_params = {}
    for k in param_dict.keys():
        if k in reserved_param_keywords: continue
        example_params[k]=param_dict[k]

    # Load the listing name module
    full_name = _full_listing_name(chapter, listing, name)
    mod = import_module(full_name)

    ex_fun = getattr(mod, name,None)
    if ex_fun is not None:
        # Run the function, passing the parameters
        ex_fun(**example_params)
    else:
        print('Could not find function %s in module %s' % (name , full_name))
        exit(-5)


def load_and_check_listing_params(args):
    """
    Loads the JSON of parameters for this schema, and gets the parameters for the specified listing. It raises errors if
    it cannot find what it is looking for. Once it finds the entry for this listing, it starts with the chapter default
    params, and then adds any parameters specific to the listing (which would override the chapter parameters, if there
    was one of the same name.) It also adds context information to the result, so that it has all the information needed
    when the listings are run.

    :param args: result of parsing arguments
    :return: dictionary of name value pairs for the example
    """
    schema = args.schema
    chapter = args.chapter
    listing = args.listing
    version = args.version

    chapter_key = f'chap{chapter}'
    list_key = f'list{listing}'
    vers_key = f'v{version}' if version else None

    # Error if there is no file for this schema
    conf_path='../../listings/conf/%s_listings.json' % schema
    if not os.path.isfile(conf_path):
        print(f'No params {conf_path} to run listings on schema {schema}')
        exit(-1)

    with open(conf_path, 'r') as myfile:
        param_dict=json.loads(myfile.read())

    # Error if there is no key for this chapter in the dictionary
    if chapter_key not in param_dict:
        print(f'No params for chapter {chapter} in {schema}_listings.json')
        exit(-2)

    # Error if there is no key for this listing
    if list_key not in param_dict[chapter_key]:
        print(f'No listing for chapter {chapter} in {schema}_listings.json')
        exit(-3)

    # Start with the chapter default parameters
    listing_params = copy(param_dict[chapter_key]['defaults'])

    # If using the insert option, set the parameter source to insert sub-block rather than regular
    if args.insert:
        if 'insert' not in param_dict[chapter_key][list_key]:
            print(f'No insert section for chapter {chapter} in {schema}_listings.json')
            exit(-4)
        param_source = param_dict[chapter_key][list_key]['insert']
        # force the right mode forn insert sqls
        listing_params['mode']='run'
    else:
        param_source = param_dict[chapter_key][list_key]

    # Add additional specific parameters for this listing, possibly overwriting defaults
    if len(param_source['params']) > 0:
        for k in param_source['params'].keys():
            listing_params[k]=param_source['params'][k]

    # add additional specific parmeters for this version (if any), possibly overwriting
    if vers_key and vers_key in param_source:
        for k in param_source[vers_key].keys():
            listing_params[k]=param_source[vers_key][k]

    # Add the contextual information
    listing_params['schema'] = schema
    listing_params['chapter']=chapter
    listing_params['listing']=listing
    listing_params['name'] = param_dict[chapter_key][list_key]['name']

    return listing_params

    # Another error if it didn't find a listing



def run_one_listing(args):
    """
    Load the dictionary of parameters for this schema
    Check the type of the listing (SQL or Python) and call the executor function
    :param args: from arg parse
    :return:
    """
    # Get arguments
    listing_params = load_and_check_listing_params(args)

    print('\nRunning chapter %d listing %d %s on schema %s' % (args.chapter,args.listing,listing_params['name'],args.schema))

    # Run the executor function for sql or for python...
    type = listing_params.get('type', listing_params['type']) # chap params should always have type
    if type=='sql':
        _sql_listing_from_params(listing_params,args.insert)
    elif type=='py':
        _python_listing_from_params(listing_params)
    else:
        raise Exception('Unsupported type %s' % type)

"""
####################################################################################################
The main script for running Fight Churn With Data examples.
This will loop over multiple listings and versions in one chapter.
"""

if __name__ == "__main__":
    args, _ = parser.parse_known_args()

    for l in args.listing:
        list_args = copy(args)
        list_args.listing = l
        if len(args.version)==0:
            list_args.version=None
            run_one_listing(list_args)
        else:
            for v in args.version:
                vers_args = copy(list_args)
                vers_args.version =v
                run_one_listing(vers_args)



import pandas as pd
from listing_7_5_fat_tail_scores import transform_fattail_columns, transform_skew_columns
from listing_8_4_rescore_metrics import score_current_data, group_current_data, reload_churn_data
from listing_10_4_dummy_variables import dummy_variables

def rescore_wcats(data_set_path,categories,groups):

    current_path = data_set_path.replace('.csv', '_current.csv')
    dummy_variables(current_path,groups, current=True)
    current_dummies = reload_churn_data(data_set_path,'current_dummies_groupscore','10.7',is_customer_data=True)
    align_dummies(current_dummies,data_set_path)

    nocat_path = data_set_path.replace('.csv', '_nocat.csv')
    load_mat_df = reload_churn_data(nocat_path,'load_mat','6.4',is_customer_data=False)
    score_df = reload_churn_data(nocat_path,'score_params','7.5',is_customer_data=False)
    current_nocat = reload_churn_data(data_set_path,'current_nocat','10.7',is_customer_data=True)
    assert set(score_df.index.values)==set(current_nocat.columns.values),"Data to re-score does not match transform params"
    assert set(load_mat_df.index.values)==set(current_nocat.columns.values),"Data to re-score does not match loading matrix"
    transform_skew_columns(current_nocat,score_df[score_df['skew_score']].index.values)
    transform_fattail_columns(current_nocat,score_df[score_df['fattail_score']].index.values)
    scaled_data = score_current_data(current_nocat,score_df,data_set_path)
    grouped_data = group_current_data(scaled_data, load_mat_df,data_set_path)

    group_dum_df = grouped_data.merge(current_dummies,left_index=True,right_index=True)
    group_dum_df.to_csv(data_set_path.replace('.csv','_current_groupscore.csv'),header=True)

    current_df = reload_churn_data(data_set_path,'current','10.7',is_customer_data=True)
    save_segment_data_wcats(grouped_data,current_df,load_mat_df,data_set_path, categories)


def align_dummies(current_data,data_set_path):
    current_groupments=pd.read_csv(data_set_path.replace('.csv','_current_dummies_groupmets.csv'),index_col=0)
    new_dummies = set(current_groupments['metrics'])
    original_groupmets = pd.read_csv(data_set_path.replace('.csv','_dummies_groupmets.csv'),index_col=0)
    old_dummies = set(original_groupmets['metrics'])
    missing_in_old = new_dummies.difference(old_dummies)
    missing_in_new = old_dummies.difference(new_dummies)
    for col in missing_in_new:
        current_data[col]=0.0
    for col in missing_in_old:
        current_data.drop(col,axis=1,inplace=True)


def save_segment_data_wcats(current_data_grouped, current_data, load_mat_df, data_set_path, categories):

    group_cols =  load_mat_df.columns[load_mat_df.astype(bool).sum(axis=0) > 1]
    no_group_cols = list(load_mat_df.columns[load_mat_df.astype(bool).sum(axis=0) == 1])
    no_group_cols.extend(categories)
    segment_df = current_data_grouped[group_cols].join(current_data[no_group_cols])
    segment_df.to_csv(data_set_path.replace('.csv','_current_groupmets_segment.csv'),header=True)



import pandas as pd
import os

def dataset_stats(data_set_path):

    assert os.path.isfile(data_set_path),'"{}" is not a valid dataset path'.format(data_set_path)
    churn_data = pd.read_csv(data_set_path,index_col=[0,1])
    if 'is_churn' in churn_data:
        churn_data['is_churn']=churn_data['is_churn'].astype(float)

    summary = churn_data.describe()
    summary = summary.transpose()

    summary['skew'] = churn_data.skew()
    summary['1%'] = churn_data.quantile(q=0.01)
    summary['99%'] = churn_data.quantile(q=0.99)
    summary['nonzero'] = churn_data.astype(bool).sum(axis=0) / churn_data.shape[0]

    summary = summary[ ['count','nonzero','mean','std','skew','min','1%','25%','50%','75%','99%','max'] ]
    summary.columns = summary.columns.str.replace("%", "pct")

    save_path = data_set_path.replace('.csv', '_summarystats.csv')
    summary.to_csv(save_path,header=True)
    print('Saving results to %s' % save_path)

